import pytest

from supriya.patterns import (
    CompositeEvent,
    EventPattern,
    GroupAllocateEvent,
    GroupPattern,
    NodeFreeEvent,
    NoteEvent,
    NullEvent,
    SequencePattern,
)
from supriya.patterns.testutils import MockUUID as M
from supriya.patterns.testutils import run_pattern_test


@pytest.mark.parametrize(
    "stop_at, inner_pattern, release_time, expected, is_infinite",
    [
        (
            None,
            EventPattern(a=SequencePattern([1, 2])),
            0.0,
            [
                CompositeEvent([GroupAllocateEvent(M("A"), delta=0.0)]),
                NoteEvent(M("B"), a=1, target_node=M("A")),
                NoteEvent(M("C"), a=2, target_node=M("A")),
                CompositeEvent([NodeFreeEvent(M("A"), delta=0.0)]),
            ],
            False,
        ),
        (
            None,
            EventPattern(a=SequencePattern([1, 2])),
            0.25,
            [
                CompositeEvent([GroupAllocateEvent(M("A"), delta=0.0)]),
                NoteEvent(M("B"), a=1, target_node=M("A")),
                NoteEvent(M("C"), a=2, target_node=M("A")),
                CompositeEvent(
                    [NullEvent(delta=0.25), NodeFreeEvent(M("A"), delta=0.0)]
                ),
            ],
            False,
        ),
    ],
)
def test(stop_at, inner_pattern, release_time, expected, is_infinite):
    pattern = GroupPattern(inner_pattern, release_time=release_time)
    run_pattern_test(pattern, expected, is_infinite, stop_at)



import asyncio
import dataclasses
import datetime
import hashlib
import platform
import subprocess
from os import PathLike
from pathlib import Path
from typing import Coroutine, Optional, Tuple, Union

from uqbar.graphs import Grapher
from uqbar.io import open_path

import supriya

from .typing import SupportsPlot, SupportsRender, SupportsRenderMemo


@dataclasses.dataclass(frozen=True)
class PlayMemo:
    """
    For renderables that need to be captured during Sphinx output, e.g. realtime buffers.

    A realtime buffer does not exist by the time Sphinx writes output, so we capture the
    state of the buffer during Sphinx's read pass as a memo, and render the memo as
    audio during the write pass.
    """

    contents: bytes
    suffix: str

    def __render__(
        self,
        output_file_path: Optional[PathLike] = None,
        render_directory_path: Optional[PathLike] = None,
        **kwargs,
    ) -> Coroutine[None, None, Tuple[Optional[Path], int]]:
        async def render_function() -> Tuple[Path, int]:
            if output_file_path is None:
                hexdigest = hashlib.sha256(self.contents).hexdigest()
                file_name = f"audio-{hexdigest}{self.suffix}"
                path = Path(render_directory_path or supriya.output_path) / file_name
            else:
                path = Path(output_file_path)
            path.write_bytes(self.contents)
            return path, 0

        return render_function()

    @classmethod
    def from_path(cls, path: Path) -> "PlayMemo":
        return cls(contents=path.read_bytes(), suffix=path.suffix)


class Player:
    ### INITIALIZER ###

    def __init__(self, renderable: SupportsRender, **kwargs) -> None:
        self.renderable = renderable
        self.render_kwargs = kwargs

    ### SPECIAL METHODS ###

    def __call__(self) -> Tuple[Optional[Path], int]:
        path, exit_code = self.render()
        if path:
            self.open_output_path(path)
        return path, exit_code

    ### PUBLIC METHODS ###

    def open_output_path(self, output_path) -> None:
        if platform.system() == "Darwin":
            subprocess.run(
                ["open", "-a", "QuickTime Player", str(output_path)], check=True
            )
        else:
            open_path(output_path)

    def render(self) -> Tuple[Optional[Path], int]:
        return render(self.renderable, **self.render_kwargs)


class Plotter:
    ### INITIALIZER ###

    def __init__(self, plottable, **kwargs):
        self.plottable = plottable
        self.plot_kwargs = kwargs

    ### SPECIAL METHODS ###

    def __call__(self):
        output_path = self.render()
        self.open_output_path(output_path)
        return output_path

    ### PUBLIC METHODS ###

    def open_output_path(self, output_path):
        open_path(output_path)

    def render(self):
        import librosa.display
        from matplotlib import pyplot

        array, sample_rate = self.plottable.__plot__()
        fig, ax = pyplot.subplots(nrows=1)
        # TODO: Drop color="blue" after upgrading librosa > 0.10.1
        #       https://github.com/librosa/librosa/issues/1763
        librosa.display.waveshow(array, sr=sample_rate, ax=ax, color="blue")
        timestamp = (
            datetime.datetime.now().isoformat().replace(".", "-").replace(":", "-")
        )
        extension = self.plot_kwargs.get("format_", "png")
        output_path = supriya.output_path / f"{timestamp}.{extension}"
        fig.savefig(output_path)
        return output_path


def graph(graphable, format_="pdf", layout="dot"):
    return Grapher(
        graphable, format_=format_, layout=layout, output_directory=supriya.output_path
    )()


def play(renderable: SupportsRender, **kwargs):
    return Player(renderable, **kwargs)()


def plot(plottable: SupportsPlot, format_="png", **kwargs):
    return Plotter(plottable, format_=format_, **kwargs)()


def render(
    renderable: Union[SupportsRender, SupportsRenderMemo],
    output_file_path: Optional[PathLike] = None,
    render_directory_path: Optional[PathLike] = None,
    **kwargs,
) -> Tuple[Optional[Path], int]:
    if isinstance(renderable, SupportsRenderMemo):
        supports_render = renderable.__render_memo__()
    else:
        supports_render = renderable
    return asyncio.run(
        supports_render.__render__(
            output_file_path=output_file_path,
            render_directory_path=render_directory_path,
            **kwargs,
        )
    )


__all__ = ["Player", "Plotter", "graph", "play", "plot", "render"]



"""
Tools for interacting with soundfiles.
"""

import asyncio
import dataclasses
import hashlib
import shlex
from os import PathLike
from pathlib import Path
from typing import Optional, Tuple

from uqbar.io import find_executable
from uqbar.strings import to_dash_case

from . import output_path


@dataclasses.dataclass(frozen=True)
class Say:
    """
    Wrapper for OSX ``say`` command.

    ::

        >>> from supriya.soundfiles import Say
        >>> say = Say("Hello World!", voice="Daniel")
        >>> supriya.play(say)  # doctest: +SKIP
    """

    text: str
    voice: Optional[str]

    ### CLASS VARIABLES ###

    _valid_voices = (
        "Alex",
        "Alice",
        "Alva",
        "Amelie",
        "Anna",
        "Carmit",
        "Damayanti",
        "Daniel",
        "Diego",
        "Ellen",
        "Fiona",
        "Fred",
        "Ioana",
        "Joana",
        "Jorge",
        "Juan",
        "Kanya",
        "Karen",
        "Kyoko",
        "Laura",
        "Lekha",
        "Luca",
        "Luciana",
        "Maged",
        "Mariska",
        "Mei-Jia",
        "Melina",
        "Milena",
        "Moira",
        "Monica",
        "Nora",
        "Paulina",
        "Samantha",
        "Sara",
        "Satu",
        "Sin-ji",
        "Tessa",
        "Thomas",
        "Ting-Ting",
        "Veena",
        "Victoria",
        "Xander",
        "Yelda",
        "Yuna",
        "Yuri",
        "Zosia",
        "Zuzana",
    )

    ### INITIALIZER ###

    def __post_init__(self):
        if self.voice and self.voice not in self._valid_voices:
            raise ValueError(self.voice)

    ### SPECIAL METHODS ###

    async def __render__(
        self,
        output_file_path: Optional[PathLike] = None,
        render_directory_path: Optional[PathLike] = None,
        **kwargs,
    ) -> Tuple[Path, int]:
        path = self._build_output_file_path(
            output_file_path=output_file_path,
            render_directory_path=render_directory_path,
        )
        if path.exists():
            return path, 0
        if find_executable("say"):
            command = ["say", "-o", str(path)]
            if self.voice:
                command.extend(["-v", self.voice])
        else:
            command = ["espeak", "-w", str(path)]
        command.append(shlex.quote(self.text))
        process = await asyncio.create_subprocess_exec(
            *command, cwd=render_directory_path
        )
        await process.communicate()
        return path, process.returncode or 0

    ### PRIVATE METHODS ###

    def _build_file_path(self):
        md5 = hashlib.sha256()
        md5.update(self.text.encode())
        if self.voice is not None:
            md5.update(self.voice.encode())
        md5 = md5.hexdigest()
        file_path = "{}-{}.aiff".format(to_dash_case(type(self).__name__), md5)
        return Path(file_path)

    def _build_output_file_path(
        self, output_file_path=None, render_directory_path=None
    ) -> Path:
        if output_file_path:
            output_file_path = Path(output_file_path).resolve()
        elif render_directory_path:
            render_directory_path = Path(render_directory_path).resolve()
            output_file_path = render_directory_path / self._build_file_path()
        else:
            output_file_path = self._build_file_path()
            render_directory_path = Path(output_path).resolve()
            output_file_path = render_directory_path / self._build_file_path()
        return output_file_path



import shutil
from pathlib import Path

import pytest

collect_ignore = ["roots"]


@pytest.fixture(scope="session")
def remove_sphinx_projects(sphinx_test_tempdir) -> None:
    # Even upon exception, remove any directory from temp area
    # which looks like a Sphinx project. This ONLY runs once.
    roots_path = Path(sphinx_test_tempdir)
    for d in roots_path.iterdir():
        if d.is_dir():
            if Path(d, "_build").exists():
                # This directory is a Sphinx project, remove it
                shutil.rmtree(str(d))


@pytest.fixture()
def rootdir(remove_sphinx_projects: None) -> Path:
    roots = Path(__file__).parent.absolute() / "roots"
    return roots



from typing import Any, Dict, Sequence, Tuple

from ..enums import CalculationRate
from ..utils import repeat_to_length
from .core import UGen, param, ugen


@ugen(ar=True, kr=True, is_input=True, is_multichannel=True)
class In(UGen):
    """
    A bus input unit generator.

    ::

        >>> supriya.ugens.In.ar(bus=0, channel_count=4)
        <In.ar()>
    """

    bus = param(0.0)


@ugen(ar=True, kr=True, is_input=True, is_multichannel=True)
class InFeedback(UGen):
    """
    A bus input unit generator.

    Reads signal from a bus with a current or one cycle old timestamp.

    ::

        >>> in_feedback = supriya.ugens.InFeedback.ar(
        ...     bus=0,
        ...     channel_count=2,
        ... )
        >>> in_feedback
        <InFeedback.ar()>
    """

    bus = param(0.0)


@ugen(ar=True, kr=True, is_multichannel=True)
class LocalIn(UGen):
    """
    A SynthDef-local bus input.

    ::

        >>> supriya.ugens.LocalIn.ar(channel_count=2)
        <LocalIn.ar()>
    """

    default = param(0.0, unexpanded=True)

    def _postprocess_kwargs(
        self,
        *,
        calculation_rate: CalculationRate,
        **kwargs,
    ) -> Tuple[CalculationRate, Dict[str, Any]]:
        default = kwargs["default"]
        if not isinstance(default, Sequence):
            default = [default]
        kwargs["default"] = list(
            repeat_to_length([float(x) for x in default], self._channel_count)
        )
        return calculation_rate, kwargs


@ugen(ar=True, kr=True, channel_count=0, fixed_channel_count=True)
class LocalOut(UGen):
    """
    A SynthDef-local bus output.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> supriya.ugens.LocalOut.ar(
        ...     source=source,
        ... )
        <LocalOut.ar()>
    """

    source = param(unexpanded=True)


@ugen(ar=True, kr=True, is_output=True, channel_count=0, fixed_channel_count=True)
class OffsetOut(UGen):
    """
    A bus output unit generator with sample-accurate timing.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> supriya.ugens.OffsetOut.ar(
        ...     bus=0,
        ...     source=source,
        ... )
        <OffsetOut.ar()>
    """

    bus = param(0)
    source = param(unexpanded=True)


@ugen(ar=True, kr=True, is_output=True, channel_count=0, fixed_channel_count=True)
class Out(UGen):
    """
    A bus output unit generator.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> supriya.ugens.Out.ar(
        ...     bus=0,
        ...     source=source,
        ... )
        <Out.ar()>
    """

    bus = param(0)
    source = param(unexpanded=True)


@ugen(ar=True, kr=True, is_output=True, channel_count=0, fixed_channel_count=True)
class ReplaceOut(UGen):
    """
    An overwriting bus output unit generator.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> supriya.ugens.ReplaceOut.ar(
        ...     bus=0,
        ...     source=source,
        ... )
        <ReplaceOut.ar()>
    """

    bus = param(0)
    source = param(unexpanded=True)


@ugen(ar=True, kr=True, is_output=True, channel_count=0, fixed_channel_count=True)
class XOut(UGen):
    """
    A cross-fading bus output unit generator.

    ::

        >>> source = supriya.ugens.WhiteNoise.ar()
        >>> xout = supriya.ugens.XOut.ar(
        ...     bus=0,
        ...     crossfade=0.5,
        ...     source=source,
        ... )
        >>> xout
        <XOut.ar()>
    """

    bus = param(0)
    crossfade = param(0.0)
    source = param(unexpanded=True)



from typing import Any, Dict, Sequence, Tuple

from .. import utils
from ..enums import CalculationRate
from ..utils import flatten
from .core import PseudoUGen, UGen, UGenScalar, UGenVector, param, ugen


class Mix(PseudoUGen):
    """
    A down-to-mono signal mixer.

    .. container:: example

        ::

            >>> from supriya.ugens import DC, Mix, SynthDefBuilder

        ::

            >>> with SynthDefBuilder() as builder:
            ...     oscillators = [DC.ar(source=1) for _ in range(5)]
            ...     mix = Mix.new(oscillators)
            ...
            >>> synthdef = builder.build(name="mix1", optimize=False)
            >>> supriya.graph(synthdef)  # doctest: +SKIP

        ::

            >>> print(synthdef)
            synthdef:
                name: mix1
                ugens:
                -   DC.ar/0:
                        source: 1.0
                -   DC.ar/1:
                        source: 1.0
                -   DC.ar/2:
                        source: 1.0
                -   DC.ar/3:
                        source: 1.0
                -   Sum4.ar:
                        input_one: DC.ar/0[0]
                        input_two: DC.ar/1[0]
                        input_three: DC.ar/2[0]
                        input_four: DC.ar/3[0]
                -   DC.ar/4:
                        source: 1.0
                -   BinaryOpUGen(ADDITION).ar:
                        left: Sum4.ar[0]
                        right: DC.ar/4[0]

    .. container:: example

        ::

            >>> with SynthDefBuilder() as builder:
            ...     oscillators = [DC.ar(source=1) for _ in range(15)]
            ...     mix = Mix.new(oscillators)
            ...
            >>> synthdef = builder.build("mix2")
            >>> supriya.graph(synthdef)  # doctest: +SKIP

        ::

            >>> print(synthdef)
            synthdef:
                name: mix2
                ugens:
                -   DC.ar/0:
                        source: 1.0
                -   DC.ar/1:
                        source: 1.0
                -   DC.ar/2:
                        source: 1.0
                -   DC.ar/3:
                        source: 1.0
                -   Sum4.ar/0:
                        input_one: DC.ar/0[0]
                        input_two: DC.ar/1[0]
                        input_three: DC.ar/2[0]
                        input_four: DC.ar/3[0]
                -   DC.ar/4:
                        source: 1.0
                -   DC.ar/5:
                        source: 1.0
                -   DC.ar/6:
                        source: 1.0
                -   DC.ar/7:
                        source: 1.0
                -   Sum4.ar/1:
                        input_one: DC.ar/4[0]
                        input_two: DC.ar/5[0]
                        input_three: DC.ar/6[0]
                        input_four: DC.ar/7[0]
                -   DC.ar/8:
                        source: 1.0
                -   DC.ar/9:
                        source: 1.0
                -   DC.ar/10:
                        source: 1.0
                -   DC.ar/11:
                        source: 1.0
                -   Sum4.ar/2:
                        input_one: DC.ar/8[0]
                        input_two: DC.ar/9[0]
                        input_three: DC.ar/10[0]
                        input_four: DC.ar/11[0]
                -   DC.ar/12:
                        source: 1.0
                -   DC.ar/13:
                        source: 1.0
                -   DC.ar/14:
                        source: 1.0
                -   Sum3.ar:
                        input_one: DC.ar/12[0]
                        input_two: DC.ar/13[0]
                        input_three: DC.ar/14[0]
                -   Sum4.ar/3:
                        input_one: Sum4.ar/0[0]
                        input_two: Sum4.ar/1[0]
                        input_three: Sum4.ar/2[0]
                        input_four: Sum3.ar[0]
    """

    ### PRIVATE METHODS ###

    ### PUBLIC METHODS ###

    @classmethod
    def new(cls, sources):
        if not isinstance(sources, Sequence):
            sources = [sources]
        sources = list(flatten(sources, terminal_types=UGenScalar))
        summed_sources = []
        for part in utils.group_by_count(sources, 4):
            if len(part) == 4:
                summed_sources.extend(
                    Sum4.new(
                        input_one=part[0],
                        input_two=part[1],
                        input_three=part[2],
                        input_four=part[3],
                    )
                )
            elif len(part) == 3:
                summed_sources.extend(
                    Sum3.new(
                        input_one=part[0],
                        input_two=part[1],
                        input_three=part[2],
                    )
                )
            elif len(part) == 2:
                summed_sources.append(part[0] + part[1])
            else:
                summed_sources.append(part[0])
        if len(summed_sources) == 1:
            return summed_sources[0]
        return Mix.new(summed_sources)

    @classmethod
    def multichannel(cls, sources, channel_count):
        """
        Segment by channel count and mix down in parallel.

        .. container:: example

            Combine panner outputs, first with first, second with second, etc.

            ::

                >>> source = supriya.ugens.SinOsc.ar(frequency=[440, 660, 880])
                >>> panner = supriya.ugens.PanAz.ar(
                ...     channel_count=4,
                ...     source=source,
                ...     position=supriya.ugens.LFNoise2.kr(),
                ... )
                >>> mix = supriya.ugens.Mix.multichannel(panner, channel_count=4)
                >>> out = supriya.ugens.Out.ar(bus=0, source=mix)
                >>> supriya.graph(out)  # doctest: +SKIP

            ::

                >>> print(out)
                synthdef:
                    name: ...
                    ugens:
                    -   SinOsc.ar/0:
                            frequency: 440.0
                            phase: 0.0
                    -   LFNoise2.kr:
                            frequency: 500.0
                    -   PanAz.ar/0:
                            source: SinOsc.ar/0[0]
                            position: LFNoise2.kr[0]
                            amplitude: 1.0
                            width: 2.0
                            orientation: 0.5
                    -   SinOsc.ar/1:
                            frequency: 660.0
                            phase: 0.0
                    -   PanAz.ar/1:
                            source: SinOsc.ar/1[0]
                            position: LFNoise2.kr[0]
                            amplitude: 1.0
                            width: 2.0
                            orientation: 0.5
                    -   SinOsc.ar/2:
                            frequency: 880.0
                            phase: 0.0
                    -   PanAz.ar/2:
                            source: SinOsc.ar/2[0]
                            position: LFNoise2.kr[0]
                            amplitude: 1.0
                            width: 2.0
                            orientation: 0.5
                    -   Sum3.ar/0:
                            input_one: PanAz.ar/0[0]
                            input_two: PanAz.ar/1[0]
                            input_three: PanAz.ar/2[0]
                    -   Sum3.ar/1:
                            input_one: PanAz.ar/0[1]
                            input_two: PanAz.ar/1[1]
                            input_three: PanAz.ar/2[1]
                    -   Sum3.ar/2:
                            input_one: PanAz.ar/0[2]
                            input_two: PanAz.ar/1[2]
                            input_three: PanAz.ar/2[2]
                    -   Sum3.ar/3:
                            input_one: PanAz.ar/0[3]
                            input_two: PanAz.ar/1[3]
                            input_three: PanAz.ar/2[3]
                    -   Out.ar:
                            bus: 0.0
                            source[0]: Sum3.ar/0[0]
                            source[1]: Sum3.ar/1[0]
                            source[2]: Sum3.ar/2[0]
                            source[3]: Sum3.ar/3[0]

            Compare with a non-multichannel mixdown:

            ::

                >>> mix = supriya.ugens.Mix.new(panner)
                >>> out = supriya.ugens.Out.ar(bus=0, source=mix)
                >>> supriya.graph(out)  # doctest: +SKIP

            ::

                >>> print(out)
                synthdef:
                    name: ...
                    ugens:
                    -   SinOsc.ar/0:
                            frequency: 440.0
                            phase: 0.0
                    -   LFNoise2.kr:
                            frequency: 500.0
                    -   PanAz.ar/0:
                            source: SinOsc.ar/0[0]
                            position: LFNoise2.kr[0]
                            amplitude: 1.0
                            width: 2.0
                            orientation: 0.5
                    -   Sum4.ar/0:
                            input_one: PanAz.ar/0[0]
                            input_two: PanAz.ar/0[1]
                            input_three: PanAz.ar/0[2]
                            input_four: PanAz.ar/0[3]
                    -   SinOsc.ar/1:
                            frequency: 660.0
                            phase: 0.0
                    -   PanAz.ar/1:
                            source: SinOsc.ar/1[0]
                            position: LFNoise2.kr[0]
                            amplitude: 1.0
                            width: 2.0
                            orientation: 0.5
                    -   Sum4.ar/1:
                            input_one: PanAz.ar/1[0]
                            input_two: PanAz.ar/1[1]
                            input_three: PanAz.ar/1[2]
                            input_four: PanAz.ar/1[3]
                    -   SinOsc.ar/2:
                            frequency: 880.0
                            phase: 0.0
                    -   PanAz.ar/2:
                            source: SinOsc.ar/2[0]
                            position: LFNoise2.kr[0]
                            amplitude: 1.0
                            width: 2.0
                            orientation: 0.5
                    -   Sum4.ar/2:
                            input_one: PanAz.ar/2[0]
                            input_two: PanAz.ar/2[1]
                            input_three: PanAz.ar/2[2]
                            input_four: PanAz.ar/2[3]
                    -   Sum3.ar:
                            input_one: Sum4.ar/0[0]
                            input_two: Sum4.ar/1[0]
                            input_three: Sum4.ar/2[0]
                    -   Out.ar:
                            bus: 0.0
                            source[0]: Sum3.ar[0]
        """
        sources = list(flatten(sources, terminal_types=UGenScalar))
        mixes, parts = [], []
        for i in range(0, len(sources), channel_count):
            parts.append(sources[i : i + channel_count])
        for columns in zip(*parts):
            mixes.append(cls.new(columns))
        return UGenVector(*mixes)


@ugen(new=True)
class MulAdd(UGen):
    """
    An Optimized multiplication / addition ugen.

    ::

        >>> source = supriya.ugens.SinOsc.ar()
        >>> mul_add = supriya.ugens.MulAdd.new(
        ...     addend=0.5,
        ...     multiplier=-1.5,
        ...     source=source,
        ... )
        >>> mul_add
        <MulAdd.ar()>
    """

    ### CLASS VARIABLES ###

    source = param()
    multiplier = param(1.0)
    addend = param(0.0)

    ### PRIVATE METHODS ###

    @classmethod
    def _new_single(
        cls,
        addend=None,
        multiplier=None,
        calculation_rate=None,
        source=None,
        special_index=None,
    ):
        def _inputs_are_valid(source, multiplier, addend):
            if CalculationRate.from_expr(source) == CalculationRate.AUDIO:
                return True
            if CalculationRate.from_expr(source) == CalculationRate.CONTROL:
                if CalculationRate.from_expr(multiplier) in (
                    CalculationRate.CONTROL,
                    CalculationRate.SCALAR,
                ):
                    if CalculationRate.from_expr(addend) in (
                        CalculationRate.CONTROL,
                        CalculationRate.SCALAR,
                    ):
                        return True
            return False

        if multiplier == 0.0:
            return addend
        minus = multiplier == -1
        no_multiplier = multiplier == 1
        no_addend = addend == 0
        if no_multiplier and no_addend:
            return source
        if minus and no_addend:
            return -source
        if no_addend:
            return source * multiplier
        if minus:
            return addend - source
        if no_multiplier:
            return source + addend
        if _inputs_are_valid(source, multiplier, addend):
            return cls(
                addend=addend,
                multiplier=multiplier,
                calculation_rate=CalculationRate.from_expr(
                    (source, multiplier, addend)
                ),
                source=source,
            )
        if _inputs_are_valid(multiplier, source, addend):
            return cls(
                addend=addend,
                multiplier=source,
                calculation_rate=CalculationRate.from_expr(
                    (multiplier, source, addend)
                ),
                source=multiplier,
            )
        return (source * multiplier) + addend


@ugen(new=True)
class Sum3(UGen):
    """
    A three-input summing unit generator.

    ::

        >>> input_one = supriya.ugens.SinOsc.ar()
        >>> input_two = supriya.ugens.SinOsc.ar(phase=0.1)
        >>> input_three = supriya.ugens.SinOsc.ar(phase=0.2)
        >>> supriya.ugens.Sum3.new(
        ...     input_one=input_one,
        ...     input_two=input_two,
        ...     input_three=input_three,
        ... )
        <Sum3.ar()>
    """

    input_one = param()
    input_two = param()
    input_three = param()

    @classmethod
    def _new_single(cls, *, input_one, input_two, input_three, **kwargs):
        if input_three == 0:
            ugen = input_one + input_two
        elif input_two == 0:
            ugen = input_one + input_three
        elif input_one == 0:
            ugen = input_two + input_three
        else:
            ugen = cls(
                calculation_rate=None,
                input_one=input_one,
                input_two=input_two,
                input_three=input_three,
            )
        return ugen

    def _postprocess_kwargs(
        self, *, calculation_rate: CalculationRate, **kwargs
    ) -> Tuple[CalculationRate, Dict[str, Any]]:
        inputs = sorted(
            [kwargs["input_one"], kwargs["input_two"], kwargs["input_three"]],
            key=lambda x: CalculationRate.from_expr(x),
            reverse=True,
        )
        calculation_rate = CalculationRate.from_expr(inputs)
        kwargs.update(
            input_one=inputs[0],
            input_two=inputs[1],
            input_three=inputs[2],
        )
        return calculation_rate, kwargs


@ugen(new=True)
class Sum4(UGen):
    """
    A four-input summing unit generator.

    ::

        >>> input_one = supriya.ugens.SinOsc.ar()
        >>> input_two = supriya.ugens.SinOsc.ar(phase=0.1)
        >>> input_three = supriya.ugens.SinOsc.ar(phase=0.2)
        >>> input_four = supriya.ugens.SinOsc.ar(phase=0.3)
        >>> supriya.ugens.Sum4.new(
        ...     input_one=input_one,
        ...     input_two=input_two,
        ...     input_three=input_three,
        ...     input_four=input_four,
        ... )
        <Sum4.ar()>
    """

    input_one = param()
    input_two = param()
    input_three = param()
    input_four = param()

    @classmethod
    def _new_single(cls, *, input_one, input_two, input_three, input_four, **kwargs):
        if input_one == 0:
            ugen = Sum3._new_single(
                input_one=input_two, input_two=input_three, input_three=input_four
            )
        elif input_two == 0:
            ugen = Sum3._new_single(
                input_one=input_one, input_two=input_three, input_three=input_four
            )
        elif input_three == 0:
            ugen = Sum3._new_single(
                input_one=input_one, input_two=input_two, input_three=input_four
            )
        elif input_four == 0:
            ugen = Sum3._new_single(
                input_one=input_one, input_two=input_two, input_three=input_three
            )
        else:
            ugen = cls(
                calculation_rate=None,
                input_one=input_one,
                input_two=input_two,
                input_three=input_three,
                input_four=input_four,
            )
        return ugen

    def _postprocess_kwargs(
        self, *, calculation_rate: CalculationRate, **kwargs
    ) -> Tuple[CalculationRate, Dict[str, Any]]:
        inputs = sorted(
            [
                kwargs["input_one"],
                kwargs["input_two"],
                kwargs["input_three"],
                kwargs["input_four"],
            ],
            key=lambda x: CalculationRate.from_expr(x),
            reverse=True,
        )
        calculation_rate = CalculationRate.from_expr(inputs)
        kwargs.update(
            input_one=inputs[0],
            input_two=inputs[1],
            input_three=inputs[2],
            input_four=inputs[3],
        )
        return calculation_rate, kwargs



from supriya.contexts.allocators import BlockAllocator


def test_allocate():
    allocator = BlockAllocator(heap_minimum=0, heap_maximum=16)
    assert allocator.allocate(4) == 0
    assert allocator.allocate(4) == 4
    assert allocator.allocate(4) == 8
    assert allocator.allocate(4) == 12
    assert allocator.allocate(4) is None
    allocator.free(0)
    allocator.free(4)
    allocator.free(8)
    allocator.free(12)
    assert allocator.allocate(20) is None
    assert allocator.allocate_at(4, 2) == 4
    assert allocator.allocate(4) == 0
    assert allocator.allocate(4) == 6
    allocator.free(4)
    assert allocator.allocate(1) == 4
    assert allocator.allocate(1) == 5


def test_allocate_block_within_block():
    allocator = BlockAllocator()
    assert allocator.allocate_at(0, 3) == 0
    assert allocator.allocate_at(0, 2) is None
    assert allocator.allocate_at(1, 2) is None
    assert allocator.allocate_at(2, 2) is None
    assert allocator.allocate_at(6, 3) == 6
    assert allocator.allocate_at(5, 2) is None
    assert allocator.allocate_at(2, 5) is None
    assert allocator.allocate_at(0, 9) is None
    assert allocator.allocate_at(3, 3) == 3
    assert allocator.free(99) is None



import asyncio
import logging

import pytest
import pytest_asyncio
from uqbar.strings import normalize

from supriya import default
from supriya.assets.synthdefs import test_two_voice
from supriya.contexts.realtime import AsyncServer, Server
from supriya.contexts.responses import NodeInfo
from supriya.enums import NodeAction
from supriya.osc import OscBundle, OscMessage


async def get(x):
    if asyncio.iscoroutine(x):
        return await x
    return x


@pytest.fixture(autouse=True)
def use_caplog(caplog):
    caplog.set_level(logging.INFO)


@pytest_asyncio.fixture(autouse=True, params=[AsyncServer, Server])
async def context(request):
    context = request.param()
    await get(context.boot())
    context.add_synthdefs(default)
    await get(context.sync())
    yield context


@pytest.mark.asyncio
async def test_Group_children(context):
    # root node's child(ren) is the default grop
    assert context.root_node.children == [context.default_group]
    # the default group initializes without children
    assert context.default_group.children == []
    # groups initially have no children
    group = context.add_group()
    assert context.default_group.children == []  # waiting for the /n_go response
    assert group.children == []  # waiting for the /n_go response
    await get(context.sync())
    assert context.default_group.children == [group]
    assert group.children == []
    # groups can have children
    synth_a = group.add_synth(default)
    assert group.children == []  # waiting for the /n_go response
    await get(context.sync())
    assert group.children == [synth_a]
    # modifying the group's children on the context modifies in the client
    synth_b = group.add_synth(default, add_action="ADD_TO_TAIL")
    synth_c = group.add_synth(default, add_action="ADD_TO_HEAD")
    synth_d = synth_b.add_synth(default, add_action="ADD_BEFORE")
    assert group.children == [synth_a]  # waiting for the /n_go response
    await get(context.sync())
    assert group.children == [synth_c, synth_a, synth_d, synth_b]
    # freed groups have no children
    group.free()
    assert group.children == [synth_c, synth_a, synth_d, synth_b]  # waiting for /n_end
    await get(context.sync())
    assert group.children == []
    # no children on an unbooted context
    await get(context.quit())
    assert context.root_node.children == []


@pytest.mark.asyncio
async def test_Node_active(context):
    # nodes are active by default
    group = context.add_group()
    assert group.active  # waiting for the /n_go response
    await get(context.sync())
    assert group.active
    # paused nodes are inactive
    group.pause()
    assert group.active  # waiting for the /n_run response
    await get(context.sync())
    assert not group.active
    # unpaused nodes are active
    group.unpause()
    assert not group.active  # waiting for the /n_run response
    await get(context.sync())
    assert group.active


@pytest.mark.asyncio
async def test_Node_allocated(context):
    # groups can be allocated
    group = context.add_group()
    assert not group.allocated  # waiting for the /n_go response
    await get(context.sync())
    assert group.allocated
    # freed nodes are unallocated
    group.free()
    assert group.allocated  # waiting for the /n_end response
    await get(context.sync())
    assert not group.allocated
    # synths can be allocated
    synth = context.add_synth(default)
    assert not synth.allocated  # waiting for the /n_go response
    await get(context.sync())
    assert synth.allocated
    # nothing is allocated on an unbooted context
    await get(context.quit())
    assert not synth.allocated
    # allocation state is stored in the context, not the node
    await get(context.boot())
    assert not synth.allocated


@pytest.mark.asyncio
async def test_Node_parent(context):
    group = context.add_group()
    assert group.parent is None  # waiting for the /n_go response
    await asyncio.sleep(0.1)
    assert group.parent == context.default_group
    assert group.parent.parent == context.root_node


@pytest.mark.asyncio
async def test_add_group(context):
    with context.osc_protocol.capture() as transcript:
        # /g_new
        group = context.add_group()
        # /p_new
        context.add_group(parallel=True)
        with context.at(1.23):
            context.add_group(parallel=True, target_node=group)
            context.add_group(parallel=True, target_node=group)
            context.add_group(target_node=group)
            context.add_group(target_node=group)
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/g_new", 1000, 0, 1),
        OscMessage("/p_new", 1001, 0, 1),
        OscBundle(
            contents=(
                OscMessage("/p_new", 1002, 0, 1000, 1003, 0, 1000),
                OscMessage("/g_new", 1004, 0, 1000, 1005, 0, 1000),
            ),
            timestamp=1.23 + context.latency,
        ),
    ]
    # guard against invalid add-actions
    synth = context.add_synth(default)
    with pytest.raises(ValueError):
        synth.add_group(add_action="ADD_TO_HEAD")
    with pytest.raises(ValueError):
        synth.add_group(add_action="ADD_TO_TAIL")


@pytest.mark.asyncio
async def test_add_synth(context):
    with context.osc_protocol.capture() as transcript:
        bus_a = context.add_bus("AUDIO")
        bus_c = context.add_bus("CONTROL")
        synth = context.add_synth(default)
        context.add_synth(default, frequency=bus_a, amplitude="c0", pan=0.25, out=0)
        context.add_synth(test_two_voice, frequencies=(123, 456))
        with context.at(1.23):
            context.add_synth(default, add_action="ADD_AFTER", target_node=synth)
            context.add_synth(
                default, frequency=bus_c, amplitude="a16", pan=0.25, out=0
            )
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/s_new", "default", 1000, 0, 1),
        OscMessage(
            "/s_new",
            "default",
            1001,
            0,
            1,
            "amplitude",
            "c0",
            "frequency",
            "a16",
            "pan",
            0.25,
        ),
        OscMessage(
            "/s_new",
            "test_two_voice",
            1002,
            0,
            1,
            "frequencies",
            (123.0, 456.0),
        ),
        OscBundle(
            contents=(
                OscMessage("/s_new", "default", 1003, 3, 1000),
                OscMessage(
                    "/s_new",
                    "default",
                    1004,
                    0,
                    1,
                    "amplitude",
                    "a16",
                    "frequency",
                    "c0",
                    "pan",
                    0.25,
                ),
            ),
            timestamp=1.23 + context.latency,
        ),
    ]
    # guard against invalid add-actions
    synth = context.add_synth(default)
    with pytest.raises(ValueError):
        synth.add_synth(default, add_action="ADD_TO_HEAD")
    with pytest.raises(ValueError):
        synth.add_synth(default, add_action="ADD_TO_TAIL")


@pytest.mark.asyncio
async def test_free_group_children(context):
    grandparent = context.add_group()
    # setup
    parent = grandparent.add_group()
    grandparent.add_synth(default)
    parent.add_synth(default)
    assert str(await get(context.query_tree())) == normalize(
        """
        NODE TREE 0 group
            1 group
                1000 group
                    1002 default
                        amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5, out: 0.0
                    1001 group
                        1003 default
                            amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5, out: 0.0
        """
    )
    # /g_freeAll
    with context.osc_protocol.capture() as transcript:
        grandparent.free_children()
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/g_freeAll", 1000)
    ]
    assert str(await get(context.query_tree())) == normalize(
        """
        NODE TREE 0 group
            1 group
                1000 group
        """
    )
    # setup
    parent = grandparent.add_group()
    grandparent.add_synth(default)
    parent.add_synth(default)
    assert str(await get(context.query_tree())) == normalize(
        """
        NODE TREE 0 group
            1 group
                1000 group
                    1005 default
                        amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5, out: 0.0
                    1004 group
                        1006 default
                            amplitude: 0.1, frequency: 440.0, gate: 1.0, pan: 0.5, out: 0.0
        """
    )
    # /g_deepFree
    with context.osc_protocol.capture() as transcript:
        grandparent.free_children(synths_only=True)
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/g_deepFree", 1000)
    ]
    assert str(await get(context.query_tree())) == normalize(
        """
        NODE TREE 0 group
            1 group
                1000 group
                    1004 group
        """
    )


@pytest.mark.asyncio
async def test_free_node(context):
    group = context.add_group()
    synth = context.add_synth(default)
    with context.osc_protocol.capture() as transcript:
        group.free()
        synth.free()
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/n_free", 1000),
        OscMessage("/n_set", 1001, "gate", 0),
    ]


@pytest.mark.asyncio
async def test_get_synth_controls(context):
    with context.at():
        with context.add_synthdefs(default):
            synth = context.add_synth(
                default, frequency=432.0, amplitude=0.333, panning=0.1
            )
    assert await get(context.sync())
    controls = await get(synth.get("frequency", "amplitude"))
    assert {key: round(value, 3) for key, value in controls.items()} == {
        "frequency": 432.0,
        "amplitude": 0.333,
    }
    # unsynced
    with context.osc_protocol.capture() as transcript:
        with context.at():
            assert await get(synth.get("frequency", sync=False)) is None
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/s_get", 1000, "frequency")
    ]


@pytest.mark.asyncio
async def test_get_synth_control_range(context):
    # TBH, not sure what the semantics of this command are supposed to be.
    with context.at():
        with context.add_synthdefs(default):
            synth = context.add_synth(
                default, frequency=432.0, amplitude=0.25, panning=0.1
            )
    assert await get(context.sync())
    assert await get(synth.get_range("amplitude", 5)) == (0.25, 0.0, 0.0, 0.0, 0.0)
    # unsynced
    with context.osc_protocol.capture() as transcript:
        with context.at():
            assert await get(synth.get_range("frequency", 3, sync=False)) is None
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/s_getn", 1000, "frequency", 3)
    ]


@pytest.mark.asyncio
async def test_map_node(context):
    bus_a = context.add_bus("AUDIO")
    bus_c = context.add_bus("CONTROL")
    synth = context.add_synth(default)
    with context.osc_protocol.capture() as transcript:
        synth.map(frequency=bus_a, amplitude=bus_c, pan=None)
    assert transcript.filtered(received=False, status=False) == [
        OscBundle(
            contents=(
                OscMessage("/n_map", 1000, "amplitude", 0, "pan", -1),
                OscMessage("/n_mapa", 1000, "frequency", 16),
            )
        )
    ]


@pytest.mark.asyncio
async def test_move_node(context):
    group = context.add_group()
    synth = context.add_synth(default)
    with context.osc_protocol.capture() as transcript:
        group.move(target_node=synth, add_action="ADD_AFTER")
        synth.move(target_node=group, add_action="ADD_BEFORE")
        synth.move(target_node=group, add_action="ADD_TO_TAIL")
        synth.move(target_node=group, add_action="ADD_TO_HEAD")
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/n_after", 1000, 1001),
        OscMessage("/n_before", 1001, 1000),
        OscMessage("/g_tail", 1000, 1001),
        OscMessage("/g_head", 1000, 1001),
    ]


@pytest.mark.asyncio
async def test_order_nodes(context):
    group_a = context.add_group()
    group_b = context.add_group()
    group_c = context.add_group()
    with context.osc_protocol.capture() as transcript:
        group_a.order(group_b, group_c, add_action="ADD_TO_TAIL")
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/n_order", 1, 1000, 1001, 1002)
    ]


@pytest.mark.asyncio
async def test_pause_node(context):
    group = context.add_group()
    with context.osc_protocol.capture() as transcript:
        group.pause()
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/n_run", 1000, 0)
    ]


@pytest.mark.asyncio
async def test_query_node(context):
    group = context.add_group()
    await asyncio.sleep(0.1)
    assert await get(group.query()) == NodeInfo(
        action=NodeAction.NODE_QUERIED,
        head_id=-1,
        is_group=True,
        next_id=-1,
        node_id=1000,
        parent_id=1,
        previous_id=-1,
        tail_id=-1,
    )
    # unsynced
    with context.osc_protocol.capture() as transcript:
        with context.at():
            assert await get(group.query(sync=False)) is None
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/n_query", 1000)
    ]


@pytest.mark.asyncio
async def test_set_node(context):
    group = context.add_group()
    with context.osc_protocol.capture() as transcript:
        group.set((1, 2.3), (2, [3.4, 4.5]), foo=3.145, bar=4.5, baz=[1.23, 4.56])
    assert transcript.filtered(received=False, status=False) == [
        OscMessage(
            "/n_set",
            1000,
            1,
            2.3,
            2,
            [3.4, 4.5],
            "bar",
            4.5,
            "baz",
            [1.23, 4.56],
            "foo",
            3.145,
        )
    ]


@pytest.mark.asyncio
async def test_set_node_range(context):
    group = context.add_group()
    with context.osc_protocol.capture() as transcript:
        group.set_range((2, [3.4, 4.5]), baz=[1.23, 4.56])
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/n_setn", 1000, 2, 2, 3.4, 4.5, "baz", 2, 1.23, 4.56)
    ]


@pytest.mark.asyncio
async def test_unpause_node(context):
    group_a = context.add_group()
    group_b = context.add_group()
    group_c = context.add_group()
    with context.osc_protocol.capture() as transcript:
        group_a.unpause()
        with context.at(1.23):
            group_b.unpause()
            group_c.unpause()
    assert transcript.filtered(received=False, status=False) == [
        OscMessage("/n_run", 1000, 1),
        OscBundle(
            contents=(OscMessage("/n_run", 1001, 1, 1002, 1),),
            timestamp=1.23 + context.latency,
        ),
    ]



from ..enums import DoneAction
from .core import PseudoUGen, UGen, UGenOperable, param, ugen


@ugen(ar=True, kr=True, is_pure=True)
class APF(UGen):
    """
    An all-pass filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> apf = supriya.ugens.APF.ar(
        ...     frequency=440,
        ...     radius=0.8,
        ...     source=source,
        ... )
        >>> apf
        <APF.ar()[0]>
    """

    source = param()
    frequency = param(440.0)
    radius = param(0.8)


@ugen(ar=True, kr=True, is_pure=True)
class BPF(UGen):
    """
    A 2nd order Butterworth bandpass filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> b_p_f = supriya.ugens.BPF.ar(source=source)
        >>> b_p_f
        <BPF.ar()[0]>
    """

    source = param()
    frequency = param(440.0)
    reciprocal_of_q = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class BPZ2(UGen):
    """
    A two zero fixed midpass filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> bpz_2 = supriya.ugens.BPZ2.ar(
        ...     source=source,
        ... )
        >>> bpz_2
        <BPZ2.ar()[0]>
    """

    source = param()


@ugen(ar=True, kr=True, is_pure=True)
class BRF(UGen):
    """
    A 2nd order Butterworth band-reject filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> b_r_f = supriya.ugens.BRF.ar(source=source)
        >>> b_r_f
        <BRF.ar()[0]>
    """

    source = param()
    frequency = param(440.0)
    reciprocal_of_q = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class BRZ2(UGen):
    """
    A two zero fixed midcut filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> brz_2 = supriya.ugens.BRZ2.ar(
        ...     source=source,
        ... )
        >>> brz_2
        <BRZ2.ar()[0]>
    """

    source = param()


class Changed(PseudoUGen):
    """
    Triggers when a value changes.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> changed = supriya.ugens.Changed.ar(
        ...     source=source,
        ...     threshold=0,
        ... )
        >>> supriya.graph(changed)  # doctest: +SKIP

    ::

        >>> print(changed)
        synthdef:
            name: 39e1f9d61589c4acaaf297cc961d65e4
            ugens:
            -   In.ar:
                    bus: 0.0
            -   HPZ1.ar:
                    source: In.ar[0]
            -   UnaryOpUGen(ABSOLUTE_VALUE).ar:
                    source: HPZ1.ar[0]
            -   BinaryOpUGen(GREATER_THAN).ar:
                    left: UnaryOpUGen(ABSOLUTE_VALUE).ar[0]
                    right: 0.0
    """

    ### PUBLIC METHODS ###

    @classmethod
    def ar(cls, *, source, threshold=0) -> UGenOperable:
        """
        Constructs an audio-rate Changed.

        ::

            >>> source = supriya.ugens.In.ar(bus=0)
            >>> changed = supriya.ugens.Changed.ar(
            ...     source=source,
            ...     threshold=0,
            ... )
            >>> supriya.graph(changed)  # doctest: +SKIP

        ::

            >>> print(changed)
            synthdef:
                name: 39e1f9d61589c4acaaf297cc961d65e4
                ugens:
                -   In.ar:
                        bus: 0.0
                -   HPZ1.ar:
                        source: In.ar[0]
                -   UnaryOpUGen(ABSOLUTE_VALUE).ar:
                        source: HPZ1.ar[0]
                -   BinaryOpUGen(GREATER_THAN).ar:
                        left: UnaryOpUGen(ABSOLUTE_VALUE).ar[0]
                        right: 0.0

        Returns ugen graph.
        """
        return abs(HPZ1.ar(source=source)) > threshold

    @classmethod
    def kr(cls, *, source, threshold=0) -> UGenOperable:
        """
        Constructs a control-rate Changed.

        ::

            >>> source = supriya.ugens.In.ar(bus=0)
            >>> changed = supriya.ugens.Changed.kr(
            ...     source=source,
            ...     threshold=0,
            ... )
            >>> supriya.graph(changed)  # doctest: +SKIP

        ::

            >>> print(changed)
            synthdef:
                name: e2436271176995c6a0a5cac6d1553f8b
                ugens:
                -   In.ar:
                        bus: 0.0
                -   HPZ1.kr:
                        source: In.ar[0]
                -   UnaryOpUGen(ABSOLUTE_VALUE).kr:
                        source: HPZ1.kr[0]
                -   BinaryOpUGen(GREATER_THAN).kr:
                        left: UnaryOpUGen(ABSOLUTE_VALUE).kr[0]
                        right: 0.0

        Returns ugen graph.
        """
        return abs(HPZ1.kr(source=source)) > threshold


@ugen(ar=True, kr=True, is_pure=True)
class Decay(UGen):
    """
    A leaky signal integrator.

    ::

        >>> source = supriya.ugens.Impulse.ar()
        >>> decay = supriya.ugens.Decay.ar(
        ...     source=source,
        ... )
        >>> decay
        <Decay.ar()[0]>
    """

    source = param()
    decay_time = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class Decay2(UGen):
    """
    A leaky signal integrator.

    ::

        >>> source = supriya.ugens.Impulse.ar()
        >>> decay_2 = supriya.ugens.Decay2.ar(
        ...     source=source,
        ... )
        >>> decay_2
        <Decay2.ar()[0]>
    """

    source = param()
    attack_time = param(0.01)
    decay_time = param(1.0)


@ugen(ar=True, kr=True)
class DetectSilence(UGen):
    """
    Evaluates `done_action` when input falls below `threshold`.

    ::

        >>> source = supriya.ugens.WhiteNoise.ar()
        >>> source *= supriya.ugens.Line.kr(start=1, stop=0)
        >>> detect_silence = supriya.ugens.DetectSilence.kr(
        ...     done_action=supriya.DoneAction.FREE_SYNTH,
        ...     source=source,
        ...     threshold=0.0001,
        ...     time=1.0,
        ... )
        >>> detect_silence
        <DetectSilence.kr()[0]>
    """

    ### CLASS VARIABLES ###

    source = param()
    threshold = param(0.0001)
    time = param(0.1)
    done_action = param(DoneAction(0))


@ugen(ar=True, kr=True, is_pure=True)
class FOS(UGen):
    """
    A first order filter section.

    ::

        out(i) = (a0 * in(i)) + (a1 * in(i-1)) + (a2 * in(i-2)) + (b1 * out(i-1)) + (b2 * out(i-2))

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> fos = supriya.ugens.FOS.ar(
        ...     a_0=0,
        ...     a_1=0,
        ...     b_1=0,
        ...     source=source,
        ... )
        >>> fos
        <FOS.ar()[0]>
    """

    source = param()
    a_0 = param(0.0)
    a_1 = param(0.0)
    b_1 = param(0.0)


@ugen(ar=True, kr=True, is_pure=True)
class Formlet(UGen):
    """
    A FOF-like filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> formlet = supriya.ugens.Formlet.ar(
        ...     attack_time=1,
        ...     decay_time=1,
        ...     frequency=440,
        ...     source=source,
        ... )
        >>> formlet
        <Formlet.ar()[0]>
    """

    source = param()
    frequency = param(440.0)
    attack_time = param(1.0)
    decay_time = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class HPF(UGen):
    """
    A Highpass filter unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.HPF.ar(source=source)
        <HPF.ar()[0]>
    """

    source = param()
    frequency = param(440.0)


@ugen(ar=True, kr=True, is_pure=True)
class HPZ1(UGen):
    """
    A two point difference filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> hpz_1 = supriya.ugens.HPZ1.ar(
        ...     source=source,
        ... )
        >>> hpz_1
        <HPZ1.ar()[0]>
    """

    source = param()


@ugen(ar=True, kr=True, is_pure=True)
class HPZ2(UGen):
    """
    A two zero fixed midcut filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> hpz_2 = supriya.ugens.HPZ2.ar(
        ...     source=source,
        ... )
        >>> hpz_2
        <HPZ2.ar()[0]>
    """

    source = param()


@ugen(ar=True, kr=True, is_pure=True)
class Integrator(UGen):
    """
    A leaky integrator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> integrator = supriya.ugens.Integrator.ar(
        ...     coefficient=1,
        ...     source=source,
        ... )
        >>> integrator
        <Integrator.ar()[0]>
    """

    source = param()
    coefficient = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class Lag(UGen):
    """
    A lag generator.

    ::

        >>> source = supriya.ugens.In.kr(bus=0)
        >>> supriya.ugens.Lag.kr(
        ...     lag_time=0.5,
        ...     source=source,
        ... )
        <Lag.kr()[0]>
    """

    source = param()
    lag_time = param(0.1)


@ugen(ar=True, kr=True, is_pure=True)
class LagUD(UGen):
    """
    An up/down lag generator.

    ::

        >>> source = supriya.ugens.In.kr(bus=0)
        >>> supriya.ugens.LagUD.kr(
        ...     lag_time_down=1.25,
        ...     lag_time_up=0.5,
        ...     source=source,
        ... )
        <LagUD.kr()[0]>
    """

    source = param()
    lag_time_up = param(0.1)
    lag_time_down = param(0.1)


@ugen(ar=True, kr=True, is_pure=True)
class Lag2(UGen):
    """
    An exponential lag generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> lag_2 = supriya.ugens.Lag2.ar(
        ...     lag_time=0.1,
        ...     source=source,
        ... )
        >>> lag_2
        <Lag2.ar()[0]>
    """

    source = param()
    lag_time = param(0.1)


@ugen(ar=True, kr=True, is_pure=True)
class Lag2UD(UGen):
    """
    An up/down exponential lag generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> lag_2_ud = supriya.ugens.Lag2UD.ar(
        ...     lag_time_down=0.1,
        ...     lag_time_up=0.1,
        ...     source=source,
        ... )
        >>> lag_2_ud
        <Lag2UD.ar()[0]>
    """

    source = param()
    lag_time_up = param(0.1)
    lag_time_down = param(0.1)


@ugen(ar=True, kr=True, is_pure=True)
class Lag3(UGen):
    """
    An exponential lag generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> lag_3 = supriya.ugens.Lag3.ar(
        ...     lag_time=0.1,
        ...     source=source,
        ... )
        >>> lag_3
        <Lag3.ar()[0]>
    """

    source = param()
    lag_time = param(0.1)


@ugen(ar=True, kr=True, is_pure=True)
class Lag3UD(UGen):
    """
    An up/down exponential lag generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> lag_3_ud = supriya.ugens.Lag3UD.ar(
        ...     lag_time_down=0.1,
        ...     lag_time_up=0.1,
        ...     source=source,
        ... )
        >>> lag_3_ud
        <Lag3UD.ar()[0]>
    """

    source = param()
    lag_time_up = param(0.1)
    lag_time_down = param(0.1)


@ugen(ar=True, kr=True, is_pure=True)
class LeakDC(UGen):
    """
    A DC blocker.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> leak_d_c = supriya.ugens.LeakDC.ar(
        ...     source=source,
        ...     coefficient=0.995,
        ... )
        >>> leak_d_c
        <LeakDC.ar()[0]>
    """

    source = param()
    coefficient = param(0.995)


@ugen(ar=True, kr=True, is_pure=True)
class LPF(UGen):
    """
    A lowpass filter unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.LPF.ar(source=source)
        <LPF.ar()[0]>
    """

    source = param()
    frequency = param(440.0)


@ugen(ar=True, kr=True, is_pure=True)
class LPZ1(UGen):
    """
    A two point average filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> lpz_1 = supriya.ugens.LPZ1.ar(
        ...     source=source,
        ... )
        >>> lpz_1
        <LPZ1.ar()[0]>
    """

    source = param()


@ugen(ar=True, kr=True, is_pure=True)
class LPZ2(UGen):
    """
    A two zero fixed lowpass filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> lpz_2 = supriya.ugens.LPZ2.ar(
        ...     source=source,
        ... )
        >>> lpz_2
        <LPZ2.ar()[0]>
    """

    source = param()


@ugen(ar=True, kr=True, is_pure=True)
class Median(UGen):
    """
    A median filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> median = supriya.ugens.Median.ar(
        ...     length=3,
        ...     source=source,
        ... )
        >>> median
        <Median.ar()[0]>
    """

    length = param(3)
    source = param()


@ugen(ar=True, kr=True, is_pure=True)
class MidEQ(UGen):
    """
    A parametric filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> mid_eq = supriya.ugens.MidEQ.ar(
        ...     db=0,
        ...     frequency=440,
        ...     reciprocal_of_q=1,
        ...     source=source,
        ... )
        >>> mid_eq
        <MidEQ.ar()[0]>
    """

    source = param()
    frequency = param(440.0)
    reciprocal_of_q = param(1.0)
    db = param(0.0)


@ugen(ar=True, kr=True, is_pure=True)
class MoogFF(UGen):
    """
    A Moog VCF implementation.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> moog_ff = supriya.ugens.MoogFF.ar(
        ...     frequency=100,
        ...     gain=2,
        ...     reset=0,
        ...     source=source,
        ... )
        >>> moog_ff
        <MoogFF.ar()[0]>
    """

    source = param()
    frequency = param(100.0)
    gain = param(2.0)
    reset = param(0.0)


@ugen(ar=True, kr=True, is_pure=True)
class OnePole(UGen):
    """
    A one pole filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> one_pole = supriya.ugens.OnePole.ar(
        ...     coefficient=0.5,
        ...     source=source,
        ... )
        >>> one_pole
        <OnePole.ar()[0]>
    """

    source = param()
    coefficient = param(0.5)


@ugen(ar=True, kr=True, is_pure=True)
class OneZero(UGen):
    """
    A one zero filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> one_zero = supriya.ugens.OneZero.ar(
        ...     coefficient=0.5,
        ...     source=source,
        ... )
        >>> one_zero
        <OneZero.ar()[0]>
    """

    source = param()
    coefficient = param(0.5)


@ugen(ar=True, kr=True, is_pure=True)
class RHPF(UGen):
    """
    A resonant highpass filter unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.RLPF.ar(source=source)
        <RLPF.ar()[0]>
    """

    source = param()
    frequency = param(440.0)
    reciprocal_of_q = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class RLPF(UGen):
    """
    A resonant lowpass filter unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.RLPF.ar(source=source)
        <RLPF.ar()[0]>
    """

    source = param()
    frequency = param(440.0)
    reciprocal_of_q = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class Ramp(UGen):
    """
    Breaks a continuous signal into line segments.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> ramp = supriya.ugens.Ramp.ar(
        ...     lag_time=0.1,
        ...     source=source,
        ... )
        >>> ramp
        <Ramp.ar()[0]>
    """

    source = param()
    lag_time = param(0.1)


@ugen(ar=True, kr=True, is_pure=True)
class Ringz(UGen):
    """
    A ringing filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> ringz = supriya.ugens.Ringz.ar(
        ...     decay_time=1,
        ...     frequency=440,
        ...     source=source,
        ... )
        >>> ringz
        <Ringz.ar()[0]>
    """

    source = param()
    frequency = param(440.0)
    decay_time = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class SOS(UGen):
    """
    A second-order filter section.

    ::

        out(i) = (a0 * in(i)) + (a1 * in(i-1)) + (b1 * out(i-1))

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> sos = supriya.ugens.SOS.ar(
        ...     a_0=0,
        ...     a_1=0,
        ...     a_2=0,
        ...     b_1=0,
        ...     b_2=0,
        ...     source=source,
        ... )
        >>> sos
        <SOS.ar()[0]>
    """

    source = param()
    a_0 = param(0.0)
    a_1 = param(0.0)
    a_2 = param(0.0)
    b_1 = param(0.0)
    b_2 = param(0.0)


@ugen(ar=True, kr=True, is_pure=True)
class Slew(UGen):
    """
    A slew rate limiter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> slew = supriya.ugens.Slew.ar(
        ...     source=source,
        ...     up=1,
        ...     down=1,
        ... )
        >>> slew
        <Slew.ar()[0]>
    """

    source = param()
    up = param(1.0)
    down = param(1.0)


@ugen(ar=True, kr=True, is_pure=True)
class Slope(UGen):
    """
    Calculates slope of signal.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> slope = supriya.ugens.Slope.ar(
        ...     source=source,
        ... )
        >>> slope
        <Slope.ar()[0]>
    """

    source = param()


@ugen(ar=True, kr=True, is_pure=True)
class TwoPole(UGen):
    """
    A two pole filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> two_pole = supriya.ugens.TwoPole.ar(
        ...     frequency=440,
        ...     radius=0.8,
        ...     source=source,
        ... )
        >>> two_pole
        <TwoPole.ar()[0]>
    """

    source = param()
    frequency = param(440.0)
    radius = param(0.8)


@ugen(ar=True, kr=True, is_pure=True)
class TwoZero(UGen):
    """
    A two zero filter.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> two_zero = supriya.ugens.TwoZero.ar(
        ...     frequency=440,
        ...     radius=0.8,
        ...     source=source,
        ... )
        >>> two_zero
        <TwoZero.ar()[0]>
    """

    source = param()
    frequency = param(440.0)
    radius = param(0.8)



from .core import UGen, param, ugen


@ugen(ar=True)
class FreeVerb(UGen):
    """
    A FreeVerb reverb unit generator.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> supriya.ugens.FreeVerb.ar(
        ...     source=source,
        ... )
        <FreeVerb.ar()[0]>
    """

    source = param()
    mix = param(0.33)
    room_size = param(0.5)
    damping = param(0.5)



# flake8: noqa
import os
import platform

import pytest
from uqbar.strings import normalize

from supriya.enums import ParameterRate
from supriya.ugens import (
    Decay2,
    DelayC,
    In,
    Mix,
    Out,
    Parameter,
    SinOsc,
    SuperColliderSynthDef,
    SynthDefBuilder,
)


@pytest.fixture
def py_synthdef_01():
    with SynthDefBuilder(freq=440) as builder:
        sine = SinOsc.ar(frequency=builder["freq"])
        Out.ar(bus=0, source=sine)
    py_synthdef = builder.build("test")
    return py_synthdef


def test_parameters_01_parameters(py_synthdef_01):
    assert py_synthdef_01.indexed_parameters == [
        (
            0,
            Parameter(name="freq", rate=ParameterRate.CONTROL, value=440.0),
        ),
    ]


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_parameters_01_supriya_vs_sclang(py_synthdef_01):
    sc_synthdef = SuperColliderSynthDef(
        "test",
        r"""
        | freq = 440 |
        Out.ar(0, SinOsc.ar(freq: freq))
        """,
    )
    sc_compiled_synthdef = sc_synthdef.compile()
    py_compiled_synthdef = py_synthdef_01.compile()
    assert py_compiled_synthdef == sc_compiled_synthdef


def test_parameters_01_supriya_vs_bytes(py_synthdef_01):
    # fmt: off
    test_compiled_synthdef = bytes(
        b'SCgf'
        b'\x00\x00\x00\x02'
        b'\x00\x01'
            b'\x04test'
                b'\x00\x00\x00\x01'
                    b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x01'
                    b'C\xdc\x00\x00'
                b'\x00\x00\x00\x01'
                    b'\x04freq'
                    b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x03'
                    b'\x07Control'
                        b'\x01'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x01'
                    b'\x06SinOsc'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x00'
                            b'\x02'
                    b'\x03Out'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x00'
                b'\x00\x00',
    )
    # fmt: on
    py_compiled_synthdef = py_synthdef_01.compile()
    assert py_compiled_synthdef == test_compiled_synthdef


@pytest.fixture
def py_synthdef_02():
    with SynthDefBuilder(freq=1200, out=23) as builder:
        sine = SinOsc.ar(frequency=builder["freq"])
        Out.ar(bus=builder["out"], source=sine)
    py_synthdef = builder.build("test")
    return py_synthdef


def test_parameters_02_parameters(py_synthdef_02):
    assert py_synthdef_02.indexed_parameters == [
        (
            0,
            Parameter(name="freq", rate=ParameterRate.CONTROL, value=1200.0),
        ),
        (
            1,
            Parameter(name="out", rate=ParameterRate.CONTROL, value=23.0),
        ),
    ]


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_parameters_02_supriya_vs_sclang(py_synthdef_02):
    sc_synthdef = SuperColliderSynthDef(
        "test",
        r"""
        arg freq=1200, out=23;
        Out.ar(out, SinOsc.ar(freq: freq));
        """,
    )
    sc_compiled_synthdef = sc_synthdef.compile()
    py_compiled_synthdef = py_synthdef_02.compile()
    assert py_compiled_synthdef == sc_compiled_synthdef


def test_parameters_02_supriya_vs_bytes(py_synthdef_02):
    # fmt: off
    test_compiled_synthdef = bytes(
        b'SCgf'
        b'\x00\x00\x00\x02'
        b'\x00\x01'
            b'\x04test'
                b'\x00\x00\x00\x01'
                    b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x02'
                    b'D\x96\x00\x00'
                    b'A\xb8\x00\x00'
                b'\x00\x00\x00\x02'
                    b'\x04freq'
                        b'\x00\x00\x00\x00'
                    b'\x03out'
                        b'\x00\x00\x00\x01'
                b'\x00\x00\x00\x03'
                    b'\x07Control'
                        b'\x01'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00'
                        b'\x01'
                        b'\x01'
                    b'\x06SinOsc'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x00'
                            b'\x02'
                    b'\x03Out'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x00'
                b'\x00\x00',
    )
    # fmt: on
    py_compiled_synthdef = py_synthdef_02.compile()
    assert py_compiled_synthdef == test_compiled_synthdef


@pytest.fixture
def py_synthdef_03():
    builder = SynthDefBuilder(damping=0.1, delay_time=1.0, room_size=0.9)
    with builder:
        microphone = In.ar(bus=0)
        delay = DelayC.ar(
            source=microphone, maximum_delay_time=5.0, delay_time=builder["delay_time"]
        )
        Out.ar(bus=0, source=delay)
    py_synthdef = builder.build("test")
    return py_synthdef


def test_parameters_03_parameters(py_synthdef_03):
    """
    Multiple parameters, including unused parameters.
    """
    assert py_synthdef_03.indexed_parameters == [
        (
            0,
            Parameter(name="damping", rate=ParameterRate.CONTROL, value=0.1),
        ),
        (
            1,
            Parameter(name="delay_time", rate=ParameterRate.CONTROL, value=1.0),
        ),
        (
            2,
            Parameter(name="room_size", rate=ParameterRate.CONTROL, value=0.9),
        ),
    ]


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_parameters_03_supriya_vs_sclang(py_synthdef_03):
    sc_synthdef = SuperColliderSynthDef(
        "test",
        r"""
        | damping=0.1, delay_time=1.0, room_size=0.9 |
        Out.ar(0, DelayC.ar(In.ar(0), 5.0, delay_time))
        """,
    )
    sc_compiled_synthdef = sc_synthdef.compile()
    py_compiled_synthdef = py_synthdef_03.compile()
    assert py_compiled_synthdef == sc_compiled_synthdef


def test_parameters_03_supriya_vs_bytes(py_synthdef_03):
    # fmt: off
    test_compiled_synthdef = bytes(
        b'SCgf'
        b'\x00\x00\x00\x02'
        b'\x00\x01'
            b'\x04test'
                b'\x00\x00\x00\x02'
                    b'\x00\x00\x00\x00'
                    b'@\xa0\x00\x00'
                b'\x00\x00\x00\x03'
                    b'=\xcc\xcc\xcd'
                    b'?\x80\x00\x00'
                    b'?fff'
                b'\x00\x00\x00\x03'
                    b'\x07damping'
                        b'\x00\x00\x00\x00'
                    b'\ndelay_time'
                        b'\x00\x00\x00\x01'
                    b'\x09room_size'
                        b'\x00\x00\x00\x02'
                b'\x00\x00\x00\x04'
                    b'\x07Control'
                        b'\x01'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00\x00\x03'
                        b'\x00\x00'
                            b'\x01'
                            b'\x01'
                            b'\x01'
                    b'\x02In'
                        b'\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x00'
                            b'\x02'
                    b'\x06DelayC'
                        b'\x02'
                        b'\x00\x00\x00\x03'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x01'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x01'
                            b'\x02'
                    b'\x03Out'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                            b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x02'
                            b'\x00\x00\x00\x00'
                b'\x00\x00',
    )
    # fmt: on
    py_compiled_synthdef = py_synthdef_03.compile()
    assert py_compiled_synthdef == test_compiled_synthdef


@pytest.fixture
def py_synthdef_04():
    builder = SynthDefBuilder(
        a_phase=0.0, freq=440, i_decay_time=1.0, t_trig_a=0, t_trig_b=0
    )
    with builder:
        decay = Decay2.kr(
            source=(builder["t_trig_a"], builder["t_trig_b"]),
            attack_time=0.005,
            decay_time=builder["i_decay_time"],
        )
        sin_osc = SinOsc.ar(frequency=builder["freq"], phase=builder["a_phase"])
        enveloped_sin_osc = sin_osc * decay
        Out.ar(bus=0, source=enveloped_sin_osc)
    py_synthdef = builder.build("trigTest")
    return py_synthdef


@pytest.fixture
def py_synthdef_05():
    builder = SynthDefBuilder(amp=0.1, freqs=[300, 400])
    with builder:
        sines = SinOsc.ar(frequency=builder["freqs"])
        sines = Mix.new(sines)
        sines = sines * builder["amp"]
        Out.ar(bus=0, source=sines)
    py_synthdef = builder.build("arrayarg")
    return py_synthdef


def test_parameters_05_parameters(py_synthdef_05):
    """
    Literal array arguments.
    """
    assert py_synthdef_05.indexed_parameters == [
        (
            0,
            Parameter(name="amp", rate=ParameterRate.CONTROL, value=0.1),
        ),
        (
            1,
            Parameter(name="freqs", rate=ParameterRate.CONTROL, value=(300.0, 400.0)),
        ),
    ]


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_parameters_05_supriya_vs_sclang(py_synthdef_05):
    sc_synthdef = SuperColliderSynthDef(
        "arrayarg",
        r"""
        |
            amp = 0.1,
            freqs = #[300, 400]
        |
        var sines;
        sines = SinOsc.ar(freqs).sum;
        Out.ar(0, sines * amp);
        """,
    )
    sc_compiled_synthdef = bytes(sc_synthdef.compile())
    py_compiled_synthdef = py_synthdef_05.compile()
    assert py_compiled_synthdef == sc_compiled_synthdef


def test_parameters_05_supriya_vs_bytes(py_synthdef_05):
    # fmt: off
    test_compiled_synthdef = bytes(
        b'SCgf'
        b'\x00\x00\x00\x02'
        b'\x00\x01'
            b'\x08arrayarg'
                b'\x00\x00\x00\x01'
                    b'\x00\x00\x00\x00'
                b'\x00\x00\x00\x03'
                    b'=\xcc\xcc\xcd'
                    b'C\x96\x00\x00'
                    b'C\xc8\x00\x00'
                b'\x00\x00\x00\x02'
                    b'\x03amp'
                        b'\x00\x00\x00\x00'
                    b'\x05freqs'
                        b'\x00\x00\x00\x01'
                b'\x00\x00\x00\x06'
                    b'\x07Control'
                        b'\x01'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00\x00\x03'
                        b'\x00\x00'
                            b'\x01'
                            b'\x01'
                            b'\x01'
                    b'\x06SinOsc'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x00'
                                b'\x00\x00\x00\x01'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x00'
                            b'\x02'
                    b'\x06SinOsc'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x00'
                                b'\x00\x00\x00\x02'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x00'
                            b'\x02'
                    b'\x0cBinaryOpUGen'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x01'
                                b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x02'
                                b'\x00\x00\x00\x00'
                            b'\x02'
                    b'\x0cBinaryOpUGen'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x02'
                            b'\x00\x00\x00\x03'
                                b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x00'
                                b'\x00\x00\x00\x00'
                            b'\x02'
                    b'\x03Out'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x04'
                                b'\x00\x00\x00\x00'
                b'\x00\x00'
    )
    # fmt: on
    py_compiled_synthdef = py_synthdef_05.compile()
    assert py_compiled_synthdef == test_compiled_synthdef


@pytest.fixture
def py_synthdef_06():
    builder = SynthDefBuilder(amp=0.1, freqs=Parameter(lag=0.5, value=[300, 400]))
    with builder:
        sines = SinOsc.ar(frequency=builder["freqs"])
        sines = Mix.new(sines)
        sines = sines * builder["amp"]
        Out.ar(bus=0, source=sines)
    py_synthdef = builder.build("arrayarg")
    return py_synthdef


def test_parameters_06_parameters(py_synthdef_06):
    """
    Literal array arguments.
    """
    assert py_synthdef_06.indexed_parameters == [
        (
            0,
            Parameter(name="amp", rate=ParameterRate.CONTROL, value=0.1),
        ),
        (
            1,
            Parameter(
                lag=0.5,
                name="freqs",
                rate=ParameterRate.CONTROL,
                value=(300.0, 400.0),
            ),
        ),
    ]


@pytest.mark.skipif(platform.system() == "Windows", reason="hangs on Windows")
@pytest.mark.skipif(
    platform.system() == "Darwin" and os.environ.get("CI") == "true",
    reason="sclang hangs without QT",
)
def test_parameters_06_supriya_vs_sclang(py_synthdef_06):
    sc_synthdef = SuperColliderSynthDef(
        "arrayarg",
        r"""
        |
            amp = 0.1,
            freqs = #[300, 400]
        |
        var sines;
        sines = SinOsc.ar(freqs).sum;
        Out.ar(0, sines * amp);
        """,
        [0, 0.5],
    )
    sc_compiled_synthdef = bytes(sc_synthdef.compile())
    py_compiled_synthdef = py_synthdef_06.compile()
    assert py_compiled_synthdef == sc_compiled_synthdef


def test_parameters_06_supriya_vs_bytes(py_synthdef_06):
    # fmt: off
    test_compiled_synthdef = bytes(
        b'SCgf'
        b'\x00\x00\x00\x02'
        b'\x00\x01'
            b'\x08arrayarg'
                b'\x00\x00\x00\x02'
                    b'\x00\x00\x00\x00'
                    b'?\x00\x00\x00'
                b'\x00\x00\x00\x03'
                    b'=\xcc\xcc\xcd'
                    b'C\x96\x00\x00'
                    b'C\xc8\x00\x00'
                b'\x00\x00\x00\x02'
                    b'\x03amp'
                        b'\x00\x00\x00\x00'
                    b'\x05freqs'
                        b'\x00\x00\x00\x01'
                b'\x00\x00\x00\x06'
                    b'\nLagControl'
                        b'\x01'
                        b'\x00\x00\x00\x03'
                        b'\x00\x00\x00\x03'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x00'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x01'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x01'
                            b'\x01'
                            b'\x01'
                            b'\x01'
                    b'\x06SinOsc'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x00'
                                b'\x00\x00\x00\x01'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x00'
                            b'\x02'
                    b'\x06SinOsc'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x00'
                                b'\x00\x00\x00\x02'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x00'
                            b'\x02'
                    b'\x0cBinaryOpUGen'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x00'
                            b'\x00\x00\x00\x01'
                                b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x02'
                                b'\x00\x00\x00\x00'
                            b'\x02'
                    b'\x0cBinaryOpUGen'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x01'
                        b'\x00\x02'
                            b'\x00\x00\x00\x03'
                                b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x00'
                                b'\x00\x00\x00\x00'
                            b'\x02'
                    b'\x03Out'
                        b'\x02'
                        b'\x00\x00\x00\x02'
                        b'\x00\x00\x00\x00'
                        b'\x00\x00'
                            b'\xff\xff\xff\xff'
                                b'\x00\x00\x00\x00'
                            b'\x00\x00\x00\x04'
                                b'\x00\x00\x00\x00'
                b'\x00\x00'
    )
    # fmt: on
    py_compiled_synthdef = py_synthdef_06.compile()
    assert py_compiled_synthdef == test_compiled_synthdef


def test_parameters_07():
    builder = SynthDefBuilder(amplitude=0, bus=0, frequency=440)
    with builder:
        sine = SinOsc.ar(frequency=builder["frequency"])
        source = sine * builder["amplitude"]
        Out.ar(bus=builder["bus"], source=source)
    synthdef = builder.build(name="simple_sine")
    assert normalize(str(synthdef)) == normalize(
        """
        synthdef:
            name: simple_sine
            ugens:
            -   Control.kr:
                    amplitude: 0.0
                    bus: 0.0
                    frequency: 440.0
            -   SinOsc.ar:
                    frequency: Control.kr[2:frequency]
                    phase: 0.0
            -   BinaryOpUGen(MULTIPLICATION).ar:
                    left: SinOsc.ar[0]
                    right: Control.kr[0:amplitude]
            -   Out.ar:
                    bus: Control.kr[1:bus]
                    source[0]: BinaryOpUGen(MULTIPLICATION).ar[0]
        """
    )


def test_building_is_idempotent():
    builder = SynthDefBuilder(amplitude=0, bus=0, frequency=440)
    with builder as builder:
        sine = SinOsc.ar(frequency=builder["frequency"])
        source = sine * builder["amplitude"]
        Out.ar(bus=builder["bus"], source=source)
    synthdef_a = builder.build()
    synthdef_b = builder.build()
    synthdef_c = builder.build()
    assert normalize(str(synthdef_a)) == normalize(
        """
        synthdef:
            name: 937772273a43d21bcd7b9f096f42648a
            ugens:
            -   Control.kr:
                    amplitude: 0.0
                    bus: 0.0
                    frequency: 440.0
            -   SinOsc.ar:
                    frequency: Control.kr[2:frequency]
                    phase: 0.0
            -   BinaryOpUGen(MULTIPLICATION).ar:
                    left: SinOsc.ar[0]
                    right: Control.kr[0:amplitude]
            -   Out.ar:
                    bus: Control.kr[1:bus]
                    source[0]: BinaryOpUGen(MULTIPLICATION).ar[0]
        """
    )
    assert str(synthdef_a) == str(synthdef_b)
    assert str(synthdef_b) == str(synthdef_c)



"""
Tools for interacting with scsynth-compatible execution contexts.
"""

from .core import Context
from .entities import (
    Buffer,
    BufferGroup,
    Bus,
    BusGroup,
    ContextObject,
    Group,
    Node,
    Synth,
)
from .nonrealtime import Score
from .realtime import AsyncServer, BaseServer, Server

__all__ = [
    "AsyncServer",
    "BaseServer",
    "Buffer",
    "BufferGroup",
    "Bus",
    "BusGroup",
    "Context",
    "ContextObject",
    "Group",
    "Node",
    "Score",
    "Server",
    "Synth",
]



import dataclasses
import enum
from functools import total_ordering
from typing import Callable, Dict, Literal, NamedTuple, Optional, Tuple, Union

Quantization = Literal[
    "8M",
    "4M",
    "2M",
    "1M",
    "1/2",
    "1/2T",
    "1/4",
    "1/4T",
    "1/8",
    "1/8T",
    "1/16",
    "1/16T",
    "1/32",
    "1/32T",
    "1/64",
    "1/64T",
    "1/128",
]


class EventType(enum.IntEnum):
    CHANGE = 0
    SCHEDULE = 1


class TimeUnit(enum.IntEnum):
    BEATS = 0
    SECONDS = 1
    MEASURES = 2


class ClockState(NamedTuple):
    beats_per_minute: float
    initial_seconds: float
    previous_measure: int
    previous_offset: float
    previous_seconds: float
    previous_time_signature_change_offset: float
    time_signature: Tuple[int, int]


@dataclasses.dataclass(frozen=True)
class Moment:
    __slots__ = (
        "beats_per_minute",
        "measure",
        "measure_offset",
        "offset",
        "seconds",
        "time_signature",
    )
    beats_per_minute: float
    measure: int
    measure_offset: float
    offset: float  # the beat since zero
    seconds: float  # the seconds since zero
    time_signature: Tuple[int, int]


@dataclasses.dataclass(frozen=True)
class Action:
    event_id: int
    event_type: int


@dataclasses.dataclass(frozen=True)
class Command(Action):
    quantization: Optional[Quantization]
    schedule_at: float
    time_unit: Optional[TimeUnit]


@dataclasses.dataclass(frozen=True)
class CallbackCommand(Command):
    args: Optional[Tuple]
    kwargs: Optional[Dict]
    procedure: Callable[["ClockContext"], Union[None, float, Tuple[float, TimeUnit]]]


@dataclasses.dataclass(frozen=True)
class ChangeCommand(Command):
    beats_per_minute: Optional[float]
    time_signature: Optional[Tuple[int, int]]


@total_ordering
@dataclasses.dataclass(frozen=True, eq=False)
class Event(Action):
    seconds: float
    measure: Optional[int]
    offset: Optional[float]

    def __eq__(self, other: object) -> bool:
        # Need to act like a tuple here
        if not isinstance(other, Event):
            return NotImplemented
        return (self.seconds, self.event_type, self.event_id) == (
            other.seconds,
            other.event_type,
            other.event_id,
        )

    def __lt__(self, other: object) -> bool:
        # Need to act like a tuple here
        if not isinstance(other, Event):
            return NotImplemented
        return (self.seconds, self.event_type, self.event_id) < (
            other.seconds,
            other.event_type,
            other.event_id,
        )


@dataclasses.dataclass(frozen=True, eq=False)
class CallbackEvent(Event):
    procedure: Callable[["ClockContext"], Union[None, float, Tuple[float, TimeUnit]]]
    args: Optional[Tuple]
    kwargs: Optional[Dict]
    invocations: int

    def __hash__(self) -> int:
        return hash((type(self), self.event_id))


@dataclasses.dataclass(frozen=True, eq=False)
class ChangeEvent(Event):
    beats_per_minute: Optional[float]
    time_signature: Optional[Tuple[int, int]]

    def __hash__(self) -> int:
        return hash((type(self), self.event_id))


class ClockContext(NamedTuple):
    current_moment: Moment
    desired_moment: Moment
    event: Union[CallbackEvent, ChangeEvent]



from .core import UGen, param, ugen


@ugen(ar=True, kr=True)
class Ball(UGen):
    """
    A bouncing ball physical model.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> ball = supriya.ugens.Ball.ar(
        ...     damping=0,
        ...     friction=0.01,
        ...     gravity=1,
        ...     source=source,
        ... )
        >>> ball
        <Ball.ar()[0]>
    """

    source = param()
    gravity = param(1.0)
    damping = param(0.0)
    friction = param(0.01)


@ugen(ar=True)
class Pluck(UGen):
    """
    A Karplus-String UGen.

    ::

        >>> source = supriya.ugens.WhiteNoise.ar()
        >>> trigger = supriya.ugens.Dust.kr(density=2)
        >>> pluck = supriya.ugens.Pluck.ar(
        ...     coefficient=0.5,
        ...     decay_time=1,
        ...     delay_time=0.2,
        ...     maximum_delay_time=0.2,
        ...     source=source,
        ...     trigger=trigger,
        ... )
        >>> pluck
        <Pluck.ar()[0]>
    """

    source = param()
    trigger = param()
    maximum_delay_time = param(0.2)
    delay_time = param(0.2)
    decay_time = param(1)
    coefficient = param(0.5)


@ugen(ar=True, kr=True)
class Spring(UGen):
    """
    A resonating spring physical model.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> spring = supriya.ugens.Spring.ar(
        ...     damping=0,
        ...     source=source,
        ...     spring=1,
        ... )
        >>> spring
        <Spring.ar()[0]>
    """

    source = param()
    spring = param(1.0)
    damping = param(0.0)


@ugen(ar=True, kr=True)
class TBall(UGen):
    """
    A bouncing object physical model.

    ::

        >>> source = supriya.ugens.In.ar(bus=0)
        >>> tball = supriya.ugens.TBall.ar(
        ...     damping=0,
        ...     friction=0.01,
        ...     gravity=10,
        ...     source=source,
        ... )
        >>> tball
        <TBall.ar()[0]>
    """

    source = param()
    gravity = param(10.0)
    damping = param(0.0)
    friction = param(0.01)



from ...ugens import (
    BPF,
    HPF,
    EnvGen,
    Envelope,
    Out,
    SynthDef,
    SynthDefBuilder,
    WhiteNoise,
)


def _build_clap_synthdef() -> SynthDef:
    with SynthDefBuilder(out=0, amplitude=0.5) as builder:
        envelope_generator_one = EnvGen.ar(
            envelope=Envelope(
                amplitudes=(0, 1, 0, 1, 0, 1, 0, 1, 0),
                durations=(0.001, 0.013, 0, 0.01, 0, 0.01, 0, 0.03),
                curves=(0, -3, 0, -3, 0, -3, 0, -4),
            ),
        )
        envelope_generator_two = EnvGen.ar(
            envelope=Envelope(
                amplitudes=(0, 1, 0), durations=(0.02, 0.3), curves=(0, -4)
            ),
            done_action=2,
        )
        noise_one = (
            BPF.ar(
                source=HPF.ar(
                    source=WhiteNoise.ar(),
                    frequency=600,
                ),
                frequency=2000,
                reciprocal_of_q=3,
            )
            * envelope_generator_one
        )
        noise_two = (
            BPF.ar(
                source=HPF.ar(
                    source=WhiteNoise.ar(),
                    frequency=1000,
                ),
                frequency=1200,
                reciprocal_of_q=0.7,
            )
            * envelope_generator_two
            * 0.7
        )
        Out.ar(
            bus=builder["out"],
            source=[((noise_one + noise_two) * 2).softclip() * builder["amplitude"]]
            * 2,
        )
    return builder.build(name="clap")


clap = _build_clap_synthdef()

__all__ = ["clap"]
